'''
A module for handling the earlier BSF1 Studio e.go! script files
'''

import struct
import os.path

class Error(Exception):
    pass

def _read_string(f):
    '''
    Read a null terminated string
    '''
    # read a z string
    letters = []
    ch = f.read(1)
    while ch != '\x00' and len(ch)==1:
        letters.append(ch)
        ch = f.read(1)
    return ''.join(letters)

def _read_params(f, parameters):
    '''
    Read parameters specified by a parameter string
    into a list of parsed tokens
    '''
    param_string = parameters
    my_params = []
    for ch in param_string:
        if ch == 's':
            # read a string
            param = _read_string(f) + '\x00'
        elif ch == 'h':
            # read a 16bit word
            param = f.read(2)
        elif ch == 'i':
            # read a 32bit dword
            param = f.read(4)
        my_params.append(param)
    return tuple(my_params)

def _read_location(f):
    '''
    Read a location descriptor
    '''
    # read the string
    string = _read_string(f)
    # read the offset
    offset, = struct.unpack('<L',f.read(4))
    return string, offset

func_params = [
    1*'i', # 00 (play movie?)
    0*'i', # 01 (check movie?)
    0*'i', # 02
    0*'i', # 03
    1*'i', # 04
    1*'i', # 05
    'iis', # 06 (dummy) [normally takes no arguments, hack to make work]
    2*'i', # 07
    2*'i', # 08 (store B into &A?)
    1*'i', # 09
    0*'i', # 0a
    3*'i'+'s', # 0b
    2*'i', # 0c
    0*'i', # 0d
    0*'i', # 0e -> 0d
    0*'i', # 0f
    0*'i', # 10
    0*'i', # 11 - increment some counter?
    1*'i', # 12 - add stuff to each party member?
    0*'i', # 13
    's'+2*'i', # 14 - jump to new script and free current one?
    0*'i', # 15
    4*'i', # 16 - some event handler?
    1*'i', # 17
    1*'i', # 18
    2*'i'+'s', # 19
    0*'i', # 1a
    4*'i', # 1b
    0*'i', # 1c
    's', # 1d (open a menu?)
    1*'i'+'s', # 1e
    1*'i', # 1f
    0*'i', # 20
    2*'i', # 21
    3*'i', # 22
    1*'i', # 23
    1*'i', # 24
    2*'i', # 25 
    2*'i', # 26
    1*'i', # 27
    0*'i', # 28 (gen.dat)
    0*'i', # 29 (work?)
    1*'i', # 2a (levelup?)
    1*'i', # 2b (get item?)
    0*'i', # 2c (booboo)
    #0*'i', # 2d (something with display?)
    2*'i', # 2d (something with display?) (hack, assumes certain sequence of flags pass)
    1*'i', # 2e
    's', # 2f (transition to new script?)
    'i', # 30 floating point?
    2*'i'+'s', # 31 load sprite?
    0*'i', # 32 -> 06
    1*'i', # 33 (getkey?)
    0*'i', # 34
    0*'i', # 35 ( go to start?)
    0*'i', # 36 ( rect intersect?)
    0*'i', # 37
    2*'i'+'si', # 38 (some flowers?)
    1*'i', # 39
    2*'i', # 3a (more flower)
    0*'i', # 3b -> 06
    1*'i', # 3c (some more item/powerup?)
    0*'i', # 3d
    1*'i', # 3e
    1*'i', # 3f
    0*'i', # 40
]

def opcode00(f):
    # first, read the function number
    func_str = f.read(2)
    func_num, = struct.unpack('<H',func_str)
    if func_num >= len(func_params):
        raise Error('Invalid function %02X at location %x' % (func_num, f.tell()-2))
    # now look up the parameters
    param_string = func_params[func_num]
    # parse the param string
    return (func_str,) + _read_params(f,param_string)

def opcode04(f):
    'read a string, if it starts with +(\x2b) or *(\x2a), read two more dwords'
    # first, read the string
    s = _read_string(f) + '\x00'
    # check for special condition
    if s[0] == '\x2b' or s[0] == '\x2a':
        # read two more dwords
        d1 = f.read(4)
        d2 = f.read(4)
        return (s, d1, d2)
    else:
        return (s,)

def opcode12(f):
    'read 16bit, then read that many strings'
    # read 16bit
    arg = f.read(2)
    num_strs, = struct.unpack('<H',arg)
    # read that many strings
    result = [arg]
    for i in range(num_strs):
        result.append(_read_string(f)+'\x00')
    return tuple(result)

def opcode19(f):
    'kinda like opcode12'
    return opcode12(f)

def opcode1b(f):
    'read 2 16bit numbers, and read an extra dword if the first number was 4'
    # read 2x16bit
    arg = f.read(4)
    arg1,arg2 = struct.unpack('<HH',arg)
    if arg1 == 4:
        return (arg[0:2],arg[2:4],f.read(4))
    else:
        return (arg[0:2],arg[2:4])

opcode_params = [
    opcode00,
    's', # 01
    's', # 02
    '', # 03 (key?)
    opcode04, # 04
    's', # 05
    'h', # 06 (some display thingy?)
    'h', # 07 -> 06
    'h', # 08 -> 06
    'h', # 09 -> 06
    'h', # 0a
    's', # 0b (jump?)
    's', # 0c ( real jump?)
    's', # 0d (conditional jump if equal?)
    's', # 0e (conditional jump if not equal?)
    's', # 0f (conditional jump if equal?)
    's', # 10 (conditional jump if not equal?)
    'ss', # 11 (long jump?)
    opcode12, # 12 (choice?)
    'hh', # 13
    'hh', # 14
    'hh', # 15
    'hh', # 16
    'hh', # 17
    'hh', # 18
    opcode19, # 19
    'ss', # 1a -> 11
    opcode1b, # 1b
    'hh', # 1c
    '', # 1d
    's', # 1e
    'h', # 1f
    'h', # 20
    'hh', # 21
    's', # 22
    'sh', # 23
    'h', # 24
    's', # 25
    's', # 26 -> 05
    'shh', # 27
    '', # 28 (stop movie?)
    '', # 29 (nothing)
    '', # 2a
    '', # 2b
    '', # 2c
    'h', # 2d
    'hh', # 2e
    'hh', # 2f
    ]


def _read_instruction(f):
    '''
    Read an instruction from the file
    '''
    # read an opcode
    opcode_str = f.read(2)
    opcode_number, = struct.unpack('<H',opcode_str)
    if opcode_number >= len(opcode_params):
        # this is a normal string
        f.seek(-2,1) # go backwards
        # read the string
        s = _read_string(f)
        # check for validity
        try:
            s.decode('shift-jisx0213')
            # okay decoding
            return ('',s+'\x00')
        except Exception:
            raise Error('Invalid opcode %02X at location %x' % (opcode_number, f.tell()-2))
    # look up the parameters for the opcode
    parameters = opcode_params[opcode_number]
    try:
        # run custom handler, if there is one
        return (opcode_str,) + parameters(f)
    except TypeError:
        # no custom handler
        return (opcode_str,) + _read_params(f,parameters)
                
        
class BSF1:
    def __init__(self):
        self.clear()

    def clear(self):
        self.locations = []
        self.location_by_name = {}
        self.location_by_address = {}
        self.instructions = []

    def _add_location(self, name, address):
        i = len(self.locations)
        self.locations.append((name,address))
        self.location_by_name[name] = i
        self.location_by_address[address] = i

    def load(self, url, dump=False):
        '''
        Load a script from a .scr file
        '''
        file_size = os.path.getsize(url)
        f=open(url,'rb')
        if f.read(4) != 'BSF1':
            raise Error('Not a valid BSF1 file!')
        self.clear()
        # read the first entry
        name, first_address = _read_location(f)
        self._add_location(name,first_address)
        while f.tell() < first_address:
            # add the rest of the entries
            name, address = _read_location(f)
            self._add_location(name,address)
        # read the instructions
        line_queue = [] # list of lines read so far
        while f.tell() < file_size:
            if dump: print hex(f.tell()),
            instruction = _read_instruction(f)
            if dump: print ' '.join([x.encode('hex') for x in instruction])
            if instruction[0] != '':
                if line_queue:
                    # append the line queue
                    self.instructions.append(('',)+tuple(line_queue))
                    line_queue = []
                self.instructions.append(instruction)
            else:
                # it's a normal text instruction, add it to the line queue
                line_queue.append(instruction[1])
        if line_queue:
            # append the line queue
            self.instructions.append(('',)+tuple(line_queue))
            line_queue = []
        

    def replace(self, new_instructions):
        '''
        Replace the old instructions with a set of new instructions
        '''
        # make a lookup table of old offset -> new offset
        if len(new_instructions) != len(self.instructions):
            raise Error('Replacement instructions must be of same quantity')
        old_compacted = [''.join(x) for x in self.instructions]
        new_compacted = [''.join(x) for x in new_instructions]
        starting_offset = self.locations[0][1]
        old_offsets = reduce(lambda x,y: x+[x[-1]+len(y)],old_compacted,[starting_offset])
        new_offsets = reduce(lambda x,y: x+[x[-1]+len(y)],new_compacted,[starting_offset])
        offset_pairs = map(lambda x,y: [x,y],old_offsets,new_offsets)
        offset_map = dict(offset_pairs)
        # redo the locations
        new_locations = []
        new_location_by_name = {}
        new_location_by_address = {}
        for name, offset in self.locations:
            if offset not in offset_map:
                raise Error('Old offset '+hex(offset)+' was not a valid offset into the parsed instructions')
            new_offset = offset_map[offset]
            new_locations.append((name,new_offset))
            new_location_by_name[name] = new_offset
            new_location_by_address[new_offset] = name
        # write in the new instructions and locations
        self.instructions =  new_instructions
        self.locations = new_locations
        self.location_by_name = new_location_by_name
        self.location_by_address = new_location_by_address

    def save(self, url):
        '''
        Save a script to a new .scr file
        '''
        # write the header
        f=open(url,'wb')
        f.write('BSF1')
        # write the locations
        for name,address in self.locations:
            f.write(name+'\x00')
            f.write(struct.pack('<L',address))
        # write the instructions
        for instruction in self.instructions:
            f.write(''.join(instruction))
        f.close()
