'''
Module for loading Studio e.go! script files
'''

import bsf1
import textwrap

class Error(Exception):
    pass

def _insert_line(scriptblock, speaker, line, lineNum):
    '''
    Helper function for inserting a single line
    '''
    sbline = scriptblock.newLine()
    sbline.otext = line
    sbline.ttext = line
    sbline.lineNum = lineNum
    sbline.speaker = speaker
    scriptblock.lines.append(sbline)
    scriptblock.linesTable[lineNum] = sbline

def process_speaker(text):
    return '',text

def import_from_oldego_file(scriptblock, filepath, options):
    '''
    Import the script from a .scr file (BSF1)
    '''
    x=bsf1.BSF1()
    x.load(filepath)
    scriptblock.clear()
    scriptblock.type = 'ego'
    scriptblock.dirty = True
    # parse in the needed text lines
    cur_speaker = ''
    cur_address = '0:0'
    speakers = {}
    instruction_number = 0
    cur_offset = x.locations[0][1]
    face_enabled = False
    for instruction in x.instructions:
        if instruction[0] == '\x01\x00':
            # enable face for the next line
            face_enabled = True
        if instruction[0] == '\x02\x00':
            # change the current speaker
            cur_speaker = instruction[1][0:-1].decode('shift-jisx0213')
            cur_address = str(instruction_number) + ':1'
        elif instruction[0] == '':
            # normal script line
            line = '\\n'.join([script_line[0:-1].decode('shift-jisx0213') for script_line in instruction[1:]])
            line_id = str(instruction_number) + ':1'
            if cur_address == '0:0':
                line_speaker = '(none)'
            else:
                line_speaker = cur_speaker
                line_id += ','+cur_address
                speakers[cur_speaker] = True
            if face_enabled:
                line_id += ',-1'
                face_enabled = False
            _insert_line(scriptblock, line_speaker, line, line_id)
        elif instruction[0] == '\x12\x00':
            # choices
            choice_num = 2
            for choice in instruction[2:]:
                _insert_line(scriptblock, '(choice)', choice[0:-1].decode('shift-jisx0213'), str(instruction_number) + ':' + str(choice_num))
                choice_num += 1
        elif instruction[0] == '\x00\x00' and instruction[1] == '\x0b\x00':
            # map location
            _insert_line(scriptblock, '(location)', instruction[5][0:-1].decode('shift-jisx0213'), str(instruction_number) + ':5')
        instruction_number += 1
        cur_offset += len(''.join(instruction))

    for speaker in speakers:
        if speaker == '*':
            continue
        sbname = scriptblock.newName()
        sbname.oname = speaker
        sbname.tname = speaker
        scriptblock.namesTable[speaker] = sbname
        scriptblock.names.append(speaker)

    # create a meta variable about the type of script
    doc,scriptblock.meta = scriptblock.newMeta()
    scriptblock.meta.setAttribute('type','BSF1')

def _process_patch_locations(s):
    '''
    Evaluate the patch location lineNum into a list of tuples
    '''
    return [ tuple([int(y) for y in x.split(':')]) for x in s.split(',')]

def _patch_instruction(instructions, patch_location, s):
    '''
    Patch an instruction
    '''
    old_instruction = list(instructions[patch_location[0]])
    old_instruction[patch_location[1]] = s
    instructions[patch_location[0]] = tuple(old_instruction)

_oldego_export_options = {
    'select_files':[
    ('Please choose the original script file',
     '','',
     'Old (BSF1) Studio e.go! script file (*.scr)|*.scr',False)
    ]
    }

def export_to_oldego_file(scriptblock, filepath, options):
    '''
    Reinsert the script
    '''
    origpath = options['select_files'][0][0]
    # number of characters available per line for descriptions
    if 'desc_width' in options:
        desc_width = options['desc_width']
    else:
        desc_width = 40
    # number of characters available per line for dialogue
    if 'char_width' in options:
        char_width = options['char_width']
    else:
        char_width = 28
    # whether overflowing a box is an error condition
    if 'catch_overflow' in options:
        catch_overflow = options['catch_overflow']
    else:
        catch_overflow = False
    # whether we should throw an error at the end if there were any overflows
    if 'log_overflow' in options:
        log_overflow = options['log_overflow']
    else:
        log_overflow = True
    # load the original file
    script_file=bsf1.BSF1()
    script_file.load(origpath)
    # create text wrappers
    desc_wrapper = textwrap.TextWrapper(width=desc_width,
                                        replace_whitespace=False)
    char_wrapper = textwrap.TextWrapper(width=char_width,
                                        replace_whitespace=False)
    # patch the instructions
    instructions = script_file.instructions[:]
    overflow_lines = []
    for line in scriptblock.lines:
        # dump the patch locations
        patch_locations = _process_patch_locations(line.lineNum)
        # classify by type
        if line.speaker == '(choice)' or line.speaker == '(location)':
            # no line breaks needed for choice/location
            _patch_instruction(instructions,patch_locations[0],line.ttext.encode('shift-jisx0213') + '\x00')
        else:
            # need to break lines for regular text
            if len(patch_locations[-1]) == 1 and patch_locations[-1] == (-1,):
                # character text
                old_lines = line.ttext.split('\\n')
                wrapped_lines = ('\n'.join([char_wrapper.fill(old_line) for old_line in old_lines])).split('\n')
                if len(wrapped_lines) > 3:
                    if catch_overflow:
                        raise Error('Too many lines in one text box: %d lines for line %s:\n"%s"' % (len(wrapped_lines),line.lineNum,'\\n'.join(wrapped_lines[3:])))
                    if log_overflow:
                        overflow_lines.append((line.lineNum,len(wrapped_lines),wrapped_lines))
                # write each line properly
                new_strings = [new_line.encode('shift-jisx0213')+'\x00' for new_line in wrapped_lines]
            else:
                # description text
                old_lines = line.ttext.split('\\n')
                wrapped_lines = ('\n'.join([desc_wrapper.fill(old_line) for old_line in old_lines])).split('\n')
                if len(wrapped_lines) > 3:
                    if catch_overflow:
                        raise Error('Too many lines in one text box: %d lines for line %s:\n"%s"' % (len(wrapped_lines),line.lineNum,'\\n'.join(wrapped_lines[3:])))
                    if log_overflow:
                        overflow_lines.append((line.lineNum,len(wrapped_lines),wrapped_lines))
                # write each line properly
                new_strings = [new_line.encode('shift-jisx0213')+'\x00' for new_line in wrapped_lines]
            # insert the actual text
            if patch_locations[0][1] != 1:
                raise Error('Malformed script line encountered: %s' % line.lineNum)
            instructions[patch_locations[0][0]] = ('',) + tuple(new_strings)
            # if there's a speaker, insert that as well
            if len(patch_locations) > 1 and len(patch_locations[1]) == 2 and line.speaker != '*' and line.speaker != '(none)':
                speaker = line.speaker
                if speaker in scriptblock.namesTable:
                    speaker = scriptblock.namesTable[speaker].tname
                _patch_instruction(instructions,patch_locations[1],speaker.encode('shift-jisx0213') + '\x00')
    # insert the new instructions
    script_file.replace(instructions)
    # save to new file
    script_file.save(filepath)
    # throw an exception of a log of number of overflows
    if log_overflow and overflow_lines:
        err_str = 'The script file has been exported, but too many lines in one text box were encountered for the following:'
        for lineNum,count,wrapped_lines in overflow_lines:
            err_str += '\n    %d lines for line %s:\n"%s"' % (count,lineNum,'\\n'.join(wrapped_lines[3:]))
        raise Error(err_str)

def export_to_text_file(scriptblock, filepath, options):
    '''
    Export the script to a text file
    '''
    # open the file
    f = open(filepath,'wb')
    for line in scriptblock.lines:
        if line.lineNum >= 0:
            textline = ''
            speaker = line.speaker
            if speaker in scriptblock.namesTable:
                speaker = scriptblock.namesTable[speaker].tname
            if speaker:
                textline+=(speaker+': ')
            textline+=(line.ttext)
            if line.comment.strip():
                textline+=(' //'+line.comment.strip())
            f.write(textline+'\n')
        else:
            f.write('\n')
    f.close()

# the following must be present in each plugin file

exports=[
    ('Old (BSF1) Studio e.go! script file','.scr',export_to_oldego_file,_oldego_export_options),
    ('Text file','.txt',export_to_text_file,{}),
    ]

imports=[
    ('Old (BSF1) Studio e.go! script file','.scr',import_from_oldego_file,{}),
    ]

desc=('Studio e.go! xml script file','.ego')

if 0:
                # need to break lines for regular text
            if len(patch_locations[-1]) == 1 and patch_locations[-1] == (-1,):
                # character text
                if line.ttext[0] in u'\u300c\u300e':
                    saved_prefix = line.ttext[0]
                    real_line = line.ttext[1:]
                else:
                    saved_prefix = ''
                    real_line = line.ttext
                wrapped_lines = char_wrapper.wrap(real_line)
                if catch_overflow and len(wrapped_lines) > 3:
                    raise Error('Too many lines in one text box: %d' % len(wrapped_lines))
                # pad each line properly
                suffix = ''
                new_string = saved_prefix+''.join([x+(' '*(char_width-len(x)))+suffix for x in wrapped_lines])
            else:
                # description text
                if line.ttext[0] in u'\u300c\u300e':
                    saved_prefix = line.ttext[0]
                    real_line = line.ttext[1:]
                else:
                    saved_prefix = ''
                    real_line = line.ttext
                wrapped_lines = desc_wrapper.wrap(real_line)
                if catch_overflow and len(wrapped_lines) > 3:
                    raise Error('Too many lines in one text box: %d' % len(wrapped_lines))
                # pad each line properly
                suffix = ''
                new_string = saved_prefix+''.join([x+(' '*(desc_width-len(x)))+suffix for x in wrapped_lines])
