"""
A module for handling fixed length strings in files
"""

import struct

class Error(Exception):
    pass

class FixedReplacer:
    def __init__(self, config):
        """
        Initialize a fixed length string replacer

        config is a configuration object and must have the following members:
          data_type - a string, only 'fixed_records' supported for now
          record_length - length of a fixed record
          record_format - a tokenized struct format string
          record_strings - a list of (index, encoding) for replaceable strings
          record_id - the index of an id tag for the record
        """
        self.config = config
        if self.config.data_type != 'fixed_records':
            raise Error('Data type '+str(self.config.data_type)+' not supported')
        self.clear()

    def clear(self):
        self.strings = []

    def load(self, f, pedantic=False):
        """
        Load the strings from an opened file

        f - a file or file-like object from which to load
        pedantic - set to True to bail out in warning situations
        """
        strings = []
        if self.config.data_type == 'fixed_records':
            record_format_str = ''.join(self.config.record_format)
            field_lengths = [struct.calcsize(x) for x in self.config.record_format]
            field_offsets = reduce(lambda x,y: x+[x[-1]+y], field_lengths, [0])
            done = False
            i = 0
            while not done:
                cur_pos = f.tell()
                record = f.read(self.config.record_length)
                if len(record) < self.config.record_length:
                    if len(record) == 0 or not pedantic:
                        break
                    else:
                        raise Error('Extra '+str(len(record))+' byte(s) at end of file')
                fields = struct.unpack(''.join(self.config.record_format), record)
                cur_id = fields[self.config.record_id]
                cur_strings = [(fields[x].rstrip('\x00').decode(y),field_offsets[x]+cur_pos,field_lengths[x],i,cur_id,y) for x,y in self.config.record_strings]
                strings.extend(cur_strings)
                i+=1
            self.strings = strings
        else:
            raise Error('Data type '+str(self.config.data_type)+' not supported')
            
