'''
Gadget Trial script format plugins
'''

# Gadget Trial scripts can use escape codes to do extra font formatting
#
# %w0 makes the following text not come in slowly letter by letter
# %w1 makes the following text come in letter by letter (default)
# %b toggles bold
# %c(0xbbggrr) sets the font color (0xffffff is the default)
# %z0 makes the following text a subscript
# %z1 makes the following text normal sized (default)
# %z of any other values makes the text super tiny
# %n0 makes the font gothic
# %n1 makes the font mincho
# %h(?) unknown (%h0 default?)
# %s unknown

from kogado import *
import textwrap

def process_speaker(stringtuple):
    first = stringtuple[0]
    if first.startswith(u'%w0\u3010') and first.endswith(u'\u3011%w1'):
        speaker = first[4:-4]
        line = ''.join(stringtuple[1:])
    else:
        speaker = ''
        line = ''.join(stringtuple)
    return speaker, line

def import_from_Wps_file(scriptblock, filepath, options):
    '''
    Import the script from a .Wps file
    '''
    x=Wps('GadgetT')
    x.load(filepath)
    strings = x.dump()
    scriptblock.clear()
    scriptblock.type = 'gadgett'
    i = 0
    speakers = {}
    for stringtuple in strings:
        sbline = scriptblock.newLine()
        if len(stringtuple) > 1:
            # there is a speaker present
            speaker, line = process_speaker(stringtuple)
            if speaker:
                speakers[speaker] = True
        else:
            # no speaker present
            speaker = ''
            line = stringtuple[0]
        sbline.otext = line
        sbline.lineNum = str(i)
        sbline.speaker = speaker
        scriptblock.lines.append(sbline)
        scriptblock.linesTable[str(i)] = sbline
        i=i+1

    for speaker in speakers:
        sbname = scriptblock.newName()
        sbname.oname = speaker
        sbname.tname = speaker
        scriptblock.namesTable[speaker] = sbname
        scriptblock.names.append(speaker)

    scriptblock.dirty = True

def export_to_text_file(scriptblock, filepath, options):
    '''
    Export the script to a text file
    '''
    # open the file
    f = open(filepath,'wb')
    for line in scriptblock.lines:
        if line.lineNum >= 0:
            textline = ''
            speaker = line.speaker
            if speaker in scriptblock.namesTable:
                speaker = scriptblock.namesTable[speaker].tname
            if speaker:
                textline+=(speaker+': ')
            textline+=(line.ttext)
            if line.comment.strip():
                textline+=(' //'+line.comment.strip())
            f.write(textline+'\n')
        else:
            f.write('\n')
    f.close()

_Wps_export_options = {
    'select_files':[
    ('Please choose the original Wps file',
     '','',
     'Wps file (*.Wps)|*.Wps',False)
    ]
    }

def export_to_Wps_file(scriptblock, filepath, options):
    # Get the original Wps file loaded
    origpath = options['select_files'][0][0]
    if 'width' in options:
        width = options['width']
    else:
        width = 57
    if 'maxlines' in options:
        maxlines = options['maxlines']
    else:
        maxlines = 3
    x=Wps('GadgetT')
    x.load(origpath)
    # Reformat the script
    strings = []
    wrapper = textwrap.TextWrapper(width=width,
                                   fix_sentence_endings=False,
                                   break_long_words=True)
    for line in scriptblock.lines:
        if line.speaker:
            # dialog, need to reformat
            speaker = line.speaker
            if speaker in scriptblock.namesTable:
                speaker = scriptblock.namesTable[speaker].tname
            speaker = '%w0%b'+speaker+':%b%w1'
            # add in bold
            newline = line.ttext
            newline = newline.replace('**','[STAR]')
            newline = newline.replace('*','%b')
            newline = newline.replace('[STAR]','**')
            lines = wrapper.wrap(newline)
            strings.append(tuple([speaker]+lines))
        else:
            # normal text, just put it in
            strings.append((line.ttext,))
    # Insert the script
    x.insert(strings,{'maxlines':maxlines})
    # Save to the new Wps file
    x.save(filepath)
    

# the following must be present in each plugin file

exports=[
    ('Wps file','.Wps',export_to_Wps_file,_Wps_export_options),
    ('Text file','.txt',export_to_text_file,{}),
    ]

imports=[
    ('Wps file','.Wps',import_from_Wps_file,{}),
    ]

desc=('Gadget Trial script file','.gts')
