'''
A module for handling Kogado script files
'''

import struct

class Error(Exception):
    pass

def _opcode_82(s):
    '''
    This is a choice text box
    '''
    num_choices = ord(s[0])
    if num_choices == 0:
        raise Error('Number of choices cannot be 0')
    cur_pos = 1
    for i in range(num_choices):
        cur_pos = s.find('\x00',cur_pos)
        if cur_pos < 0:
            raise Error('Could not find end of string')
        cur_pos += 1
    # add the jump table length
    cur_pos += (num_choices+1)*4
    if cur_pos > len(s):
        raise Error('End reached before finishing reading jump table')
    return cur_pos

def _opcode_83(s):
    '''
    This is a text box with an automatic pause
    '''
    window_position = ord(s[0])
    num_lines = ord(s[1])
    cur_pos = 2
    for i in range(num_lines):
        cur_pos = s.find('\x00',cur_pos)
        if cur_pos < 0:
            raise Error('Could not find end of string')
        cur_pos += 1
    return cur_pos

def _opcode_84(s):
    '''
    This is the same as 83, but with no pause
    '''
    return _opcode_83(s)
        
# a mapping from opcode->numargs for fixed opcodes
# a mapping from opcode->func for variable opcodes
# func is a function that takes as input the start of the opcode
# and returns the actual opcode length

# op1 jump table at .text:00408021 in GadgetT.exe from Trial version
_opcode_table_GadgetT = {
    0x21:6,
    0x26:0,
    0x28:0,
    0x29:0,
    0x7c:0,
    0x7d:-1,
    0x80:4,
    0x81:4,
    0x82:_opcode_82,
    0x83:_opcode_83,
    0x84:_opcode_84,
    0x85:4,
    0x100:4,
    0x101:4,
    0x102:4,
    0x103:4,
    0x104:4,
    0x105:4,
    0x106:4,
    0x107:8,
    0x108:8,
    0x109:8,
    0x10A:8,
    0x10B:8,
    0x10C:8,
    0x10D:8,
    0x10E:8,
    0x10F:8,
    0x110:8,
    0x111:8,
    0x112:8,
    0x113:4,
    0x114:4,
    0x115:0,
    0x116:4,
    0x117:4,
    0x118:4,
    0x119:4,
    0x11A:4,
    0x11B:4,
    0x11C:8,
    0x11D:4,
    0x11E:4,
    0x11F:4,
    0x120:4,
    0x121:4,
    0x122:0,
    0x123:4,
    0x124:4,
    0x125:8,
    0x126:8,
    0x127:0,
    0x128:4,
    0x129:8, #4 in AUT?
    0x12A:4,
    0x12B:8,
    0x12C:8,
    0x12D:12,
    0x12E:4,
    0x12F:4,
    0x130:0,
    0x131:4,
    0x132:4,
    0x133:8,
    0x134:0,
    0x135:0,
    0x136:12,
    0x137:8,
    0x138:8, #0 in AUT?
    0x139:4,
    0x13A:4,
    0x13B:0,
    0x13C:0,
    0x13D:0, #4 in AUT?
    0x13E:0,
    0x13F:8, #4 in AUT?
    0x140:4,
    0x141:0,
    0x142:4,
    0x143:4,
    0x144:0, #4 in AUT?
    0x145:0,
    0x146:4, #8 in AUT?
    0x147:4, #0 in AUT?
    0x148:4,
    0x149:4,
    0x14A:4,
    0x14B:0,
    0x14C:0,
    0x14D:4, #0 in AUT?
    0x14E:4,
}

_opcode_table_AUT = {
    0x21:6,
    0x26:0,
    0x28:0,
    0x29:0,
    0x7c:0,
    0x7d:-1,
    0x80:4,
    0x81:4,
    0x82:_opcode_82,
    0x83:_opcode_83,
    0x84:_opcode_84,
    0x85:4,
    0x100:4,
    0x101:4,
    0x102:4,
    0x103:4,
    0x104:4,
    0x105:8,
    0x106:8,
    0x107:8,
    0x108:8,
    0x109:8,
    0x10A:8,
    0x10B:8,
    0x10C:8,
    0x10D:8,
    0x10E:8,
    0x10F:8,
    0x110:8,
    0x111:4,
    0x112:4,
    0x113:16,
    0x114:24,
    0x115:12,
    0x116:12,
    0x117:12,
    0x118:12,
    0x119:12,
    0x11A:12,
    0x11B:4,
    0x11C:4,
    0x11D:4,
    0x11E:4,
    0x11F:4,
    0x120:4,
    0x121:4,
    0x122:4,
    0x123:4,
    0x124:4,
    0x125:4,
    0x126:4,
    0x127:4,
    0x128:4,
    0x129:4,
    0x12A:4,
    0x12B:4,
    0x12C:4,
    0x12D:4,
    0x12E:8,
    0x12F:4,
    0x130:0,
    0x131:8,
    0x132:8,
    0x133:8,
    0x134:8,
    0x135:8,
    0x136:8,
    0x137:4,
    0x138:0,
    0x139:4,
    0x13A:4,
    0x13B:4,
    0x13C:4,
    0x13D:4,
    0x13E:4,
    0x13F:4,
    0x140:8,
    0x141:4,
    0x142:4,
    0x143:4,
    0x144:4,
    0x145:4,
    0x146:8,
    0x147:0,
    0x148:4,
    0x149:8,
    0x14A:4,
    0x14B:4,
    0x14C:4,
    0x14D:0,
    0x14E:0,
    0x14F:12,
    0x150:8,
    0x151:8,
    0x152:4,
    0x153:4,
    0x154:0,
    0x155:0,
    0x156:0,
    0x157:0,
    0x158:8,
    0x159:4,
    0x15A:4,
    0x15B:4,
    0x15C:4,
    0x15D:4,
    0x15E:4,
    0x15F:4,
    0x160:0,
    0x161:4,
    0x162:4,
    0x163:4,
    0x164:8,
    0x165:4,
    0x166:0,
    0x167:4,
    0x168:4,
    0x169:12,
    0x16A:4,
    0x16B:8,
    0x16C:8,
    0x16D:4,
    0x16E:0,
}

_opcode_tables = {
    'AUT':_opcode_table_AUT,
    'GadgetT':_opcode_table_GadgetT,
    }

def _tokenize_chunk(table,chunk):
    current_pos = 0
    tokens = []
    finished = False
    # start by getting the script index and title
    script_index = ord(chunk[0])
    current_pos += 1
    end_of_title = chunk.find('\x00',current_pos)
    if end_of_title < 0:
        raise Error('Unexpected EOF while reading title name')
    title = chunk[1:end_of_title]
    tokens = [(-1,script_index),(-2,title)]
    current_pos = end_of_title + 1
    while current_pos < len(chunk) and not finished:
        opcode, = struct.unpack('<H',chunk[current_pos:current_pos+2])
        current_pos += 2
        if opcode not in table:
            for i in range(-4,0):
                print hex(tokens[i][0]),' ',tokens[i][1].encode('hex')
            raise Error('['+hex(current_pos-2)+'] Unknown opcode: '+hex(opcode))
        numargs = table[opcode]
        if '__call__' in dir(numargs):
            # a function
            numargs = numargs(chunk[current_pos:])
        if numargs < 0:
            finished = True
            numargs = -1-numargs
        token = chunk[current_pos:current_pos+numargs]
        current_pos += numargs
        tokens.append((opcode,token))
    return tokens

def _dump_title(args):
    return [(args,)]

def _dump_82(args):
    stringlist = []
    num_choices = ord(args[0])
    if num_choices == 0:
        raise Error('Number of choices cannot be 0')
    cur_pos = 1
    for i in range(num_choices):
        string_end = args.find('\x00',cur_pos)
        if string_end < 0:
            raise Error('Could not find end of string')
        stringlist.append((args[cur_pos:string_end],))
        cur_pos = string_end + 1
    return stringlist

def _dump_83(args):
    stringlist = []
    num_lines = ord(args[1])
    if num_lines == 0:
        raise Error('Number of lines cannot be 0')
    cur_pos = 2
    for i in range(num_lines):
        string_end = args.find('\x00',cur_pos)
        if string_end < 0:
            raise Error('Could not find end of string')
        stringlist.append(args[cur_pos:string_end])
        cur_pos = string_end + 1
    return [tuple(stringlist)]

def _dump_84(args):
    return _dump_83(args)

_dump_opcodes = {
    -2:_dump_title,
    0x82:_dump_82,
    0x83:_dump_83,
    0x84:_dump_84,
}

def _dump_chunk(tokens):
    stringlist = []
    for opcode,args in tokens:
        if opcode in _dump_opcodes:
            dumper = _dump_opcodes[opcode]
            stringlist.extend(dumper(args))
    return stringlist

def _compile_chunk(tokens):
    # start by setting the script index and title
    chunk = chr(tokens[0][1]) + tokens[1][1] + '\x00'
    # go through all the tokens and recombine
    for opcode,args in tokens[2:]:
        chunk += struct.pack('<H',opcode) + args
    return chunk

def _insert_title(args, strings, options):
    newtokens = [(-2,strings[0][0])]
    return newtokens, strings[1:]

def _insert_82(args, strings, options):
    num_choices = ord(args[0])
    if num_choices == 0:
        raise Error('Number of choices cannot be 0')
    new_choices = [x[0] for x in strings[0:num_choices]]
    jump_table = args[-(num_choices+1)*4:]
    newtokens = [(0x82,args[0]+'\x00'.join(new_choices)+'\x00'+jump_table)]
    return newtokens, strings[num_choices:]

def _insert_83(args, strings, options, opcode=0x83):
    if 'maxlines' in options:
        maxlines = options['maxlines']
    else:
        maxlines = 3
    newline = strings[0]
    newtokens = []
    if len(newline) > maxlines+1:
        # need multiple windows
        newlines = []
        for i in range(1,len(newline),maxlines):
            newparagraph = [newline[0]] + list(newline[i:i+maxlines])
            newlines.append(tuple(newparagraph))
    else:
        newlines = [newline]
    for newline in newlines:
        newtokens.append((opcode,args[0]+chr(len(newline))+'\x00'.join(newline)+'\x00'))
    return newtokens, strings[1:]

def _insert_84(args, strings, options):
    return _insert_83(args, strings, options=options, opcode=0x84)

_insert_opcodes = {
    -2:_insert_title,
    0x82:_insert_82,
    0x83:_insert_83,
    0x84:_insert_84,
}

def _token_length(opcode,args):
    if opcode == -1:
        # script leading number
        return 1
    elif opcode == -2:
        # script title
        return len(args)+1
    else:
        # normal opcode
        return len(args)+2

def _insert_chunk(oldtokens, strings, options):
    newtokens = []
    mapping = {}
    old_offset = 0
    new_offset = 0
    for opcode,args in oldtokens:
        mapping[old_offset]=new_offset
        if opcode in _insert_opcodes:
            inserter = _insert_opcodes[opcode]
            inserted_tokens, strings = inserter(args, strings, options)
        else:
            inserted_tokens = [(opcode,args)]
        old_length = _token_length(opcode,args)
        new_length = 0
        for new_op,new_args in inserted_tokens:
            new_length += _token_length(new_op,new_args)
        old_offset += old_length
        new_offset += new_length
        newtokens.extend(inserted_tokens)
    return newtokens, strings, mapping

def _remap_80(opcode,args,mapper,old_offset):
    '''
    A relative jump
    '''
    # old offset is the old location of the opcode
    new_offset = mapper[old_offset] # current location of the opcode
    old_rel_offset, = struct.unpack('<l',args)
    old_abs_offset = old_rel_offset + old_offset
    new_abs_offset = mapper[old_abs_offset]
    new_rel_offset = new_abs_offset - new_offset
    new_args = struct.pack('<l',new_rel_offset)
    return opcode,new_args

def _remap_81(opcode,args,mapper,old_offset):
    return _remap_80(opcode,args,mapper,old_offset)

def _remap_82(opcode,args,mapper,old_offset):
    '''
    A choice dialog
    Uses absolute jumps
    '''
    num_choices = ord(args[0])
    old_jump_table = struct.unpack('<'+'L'*num_choices,args[-(num_choices+1)*4:])
    new_jump_table = [mapper[x] for x in old_jump_table]
    new_args = args[0:-(num_choices+1)*4] + struct.pack('<'+'L'*num_choices,
                                                        *new_jump_table)
    return opcode,new_args

def _remap_85(opcode,args,mapper,old_offset):
    return _remap_80(opcode,args,mapper,old_offset)

_remap_opcodes = {
    0x80:_remap_80,
    0x81:_remap_81,
    0x82:_remap_82,
    0x85:_remap_85,
}

def _remap_chunk(tokens, mapping):
    old_offsets = mapping.keys()
    old_offsets.sort()
    for i in range(len(tokens)):
        opcode,args = tokens[i]
        if opcode in _remap_opcodes:
            remapper = _remap_opcodes[opcode]
            newop,newargs = remapper(opcode,args,mapper,old_offsets[i])
            assert(len(newargs) == len(args))
            tokens[i]=newop,newargs

class Wps:
    '''
    A class for handling .Wps files
    Takes one argument specifying type of .Wps file:
       AUT
       GadgetT
    '''
    def __init__(self, type='AUT'):
        self.clear()
        self.table = _opcode_tables[type]

    def clear(self):
        # each chunk is a list of strings each of which are opcodes+args
        self.chunks = []

    def load(self, path):
        # load a script
        self.clear()
        f = open(path,'rb')
        s=f.read()
        f.close()
        if s[1:4] != '\x00\x00\x00':
            raise Error('Not a .Wps file')
        self.lead_byte = ord(s[0])
        num_chunks, = struct.unpack('<L',s[4:8])
        chunk_offsets = struct.unpack('<'+'L'*num_chunks,s[8:8+4*num_chunks])
        for chunk_offset in chunk_offsets:
            tokens = _tokenize_chunk(self.table,s[chunk_offset:])
            self.chunks.append(tokens)

    def save(self, path):
        # save a script
        data_start = (2+len(self.chunks))*4
        cur_offset = data_start
        chunk_data=[]
        # compile the chunks
        for tokens in self.chunks:
            chunk = _compile_chunk(tokens)
            offset = cur_offset
            cur_offset += len(chunk)
            chunk_data.append((chunk,offset))
        # compile the header
        header = chr(self.lead_byte)+'\x00\x00\x00'+struct.pack('<L',len(self.chunks))+struct.pack('<'+'L'*len(self.chunks),*[x[1] for x in chunk_data])
        # write the file
        f=open(path,'wb')
        f.write(header)
        for chunk,offset in chunk_data:
            f.write(chunk)
        f.close()

    def dump(self):
        # extract a list of unicode strings
        stringlist = []
        for chunk in self.chunks:
            stringlist.extend(_dump_chunk(chunk))
        realstringlist = []
        for stringtuple in stringlist:
            realstringlist.append(tuple([x.decode('shift_jisx0213') for x in stringtuple]))
        return realstringlist

    def insert(self, unicode_strings, options):
        # insert a list of unicode strings
        newchunks = []
        strings = []
        for stringtuple in unicode_strings:
            strings.append(tuple([x.encode('shift_jisx0213') for x in stringtuple]))
        for oldchunk in self.chunks:
            # insert the strings and pop them off the list
            newchunk,strings,mapping = _insert_chunk(oldchunk, strings, options)
            # re-map offsets in the script
            _remap_chunk(newchunk,mapping)
            # add it to new chunks
            newchunks.append(newchunk)
        # set the new chunks
        self.chunks = newchunks

if __name__ == '__main__':
    # for debugging try loading a file
    import sys
    try:
        import psyco
        psyco.full()
    except ImportError:
        pass
    if len(sys.argv) > 1:
        x=Wps('GadgetT')
        x.load(sys.argv[1])
        print 'Success!'
    if len(sys.argv) > 2:
        x.save(sys.argv[2])
        print 'Success!'
    if len(sys.argv) > 3:
        f=open(sys.argv[3],'wb')
        strings = x.dump()
        for stringtuple in strings:
            for string in stringtuple:
                f.write(string.encode('utf-8'))
                f.write('\r\n'.encode('utf-8'))
        f.close()
        print 'Success!'
