"""
A module to handle script files from MarbleEngine
"""

import textwrap
import collections

class Error(Exception):
    pass

class MBLScript:
    def __init__(self, s):
        """
        Load an initial script file from string s
        """
        if s[-1] != '\x00':
            raise Error('unterminated script file')
        self.lines = s[0:-1].split('\x00')

    def dump(self):
        """
        Return a list of (speaker,line) tuples
        from the script
        """
        result = []
        cur_speaker = ''
        cur_dialogue = []

        for line in self.lines:
            if line[0:8] == '@stitle ':
                if cur_dialogue:
                    result.append((cur_speaker, '\\n'.join(cur_dialogue)))
                    cur_dialogue = []
                    cur_speaker = ''
                result.append(('title',line[8:].decode('cp932')))
            elif line[0:1] == '>':
                if cur_dialogue:
                    result.append((cur_speaker, '\\n'.join(cur_dialogue)))
                    cur_dialogue = []
                    cur_speaker = ''
                cur_speaker = line[1:].decode('cp932')
            elif line[0:1] == '{' and len(line)>1:
                if cur_dialogue:
                    result.append((cur_speaker, '\\n'.join(cur_dialogue)))
                    cur_dialogue = []
                    cur_speaker = ''
                result.append(('choice',line[1:].decode('cp932')))
            elif line[0:1] in ['*','!','@',':','}','\\','=','{']:
                if cur_dialogue:
                    result.append((cur_speaker, '\\n'.join(cur_dialogue)))
                    cur_dialogue = []
                    cur_speaker = ''
            else:
                # this is dialogue
                cur_dialogue.append(line.decode('cp932'))
                
        if cur_dialogue:
            result.append((cur_speaker, '\\n'.join(cur_dialogue)))
            cur_dialogue = []
            cur_speaker = ''

        return result

    def insert(self, new_lines, cols=44, rows=3, pedantic=True):
        """
        Insert a list of (speaker,line) tuples
        into the script by replacing what was
        originally there
        """
        current = collections.deque(self.lines)
        output = []
        cur_rows = rows

        wrapper = textwrap.TextWrapper(width=cols,
                                       fix_sentence_endings=False,
                                       break_long_words=True)

        def is_newpage(s):
            return s.startswith('\\') or s.startswith('@select')

        for (speaker, ttext, otext) in new_lines:
            _otext = otext.encode('cp932')
            _ttext = ttext.encode('cp932')
            _speaker = speaker.encode('cp932')
            
            if speaker == 'raw':
                if current and _otext == current[0]:
                    # replacement
                    current.popleft()
                    output.append(_ttext)
                    if is_newpage(_ttext):
                        # newpage
                        cur_rows = rows
                elif _otext == '':
                    # insertion
                    output.append(_ttext)
                    if is_newpage(_ttext):
                        # newpage
                        cur_rows = rows
                else:
                    if current:
                        raise Error('script mismatch, expecting raw line %s but got %s instead' % (repr(_ttext), repr(current[0])))
                    else:
                        raise Error('script mismatch, expecting raw line %s but got end of file instead' % repr(_ttext))
            elif speaker == 'choice':
                while current and current[0][0:1] != '{':
                    cur = current.popleft()
                    if is_newpage(cur):
                        # newpage
                        cur_rows = rows
                    output.append(cur)
                if current:
                    cur = current.popleft()
                    if (not pedantic) or cur == '{'+_otext:
                        output.append('{'+_ttext)
                    else:
                        raise Error('script mismatch, expecting %s but got %s instead' % (repr('{'+_otext), repr(cur)))
                else:
                    raise Error('script mismatch, expecting a choice line but hit end of script file')
            elif speaker == 'title':
                while current and current[0][0:8] != '@stitle ':
                    cur = current.popleft()
                    if is_newpage(cur):
                        # newpage
                        cur_rows = rows
                    output.append(cur)
                if current:
                    cur = current.popleft()
                    if (not pedantic) or cur == '@stitle '+_otext:
                        output.append('@stitle '+_ttext)
                    else:
                        raise Error('script mismatch, expecting %s but got %s instead' % (repr('@stitle '+_otext), repr(cur)))
                else:
                    raise Error('script mismatch, expecting a scene title line but hit end of script file')
            else:
                # do dialogue substitution
                while current and current[0][0:1] in ['>','*','!','@',':','{','}','\\','=']:
                    cur = current.popleft()
                    if cur[0:1] == '>':
                        # replace speaker
                        cur = '>'+_speaker
                    elif is_newpage(cur):
                        # newpage
                        cur_rows = rows
                    output.append(cur)
                # now current should start with a series of dialogue lines to consume
                old_lines = _otext.split('\\n')
                while current and current[0][0:1] not in ['>','*','!','@',':','{','}','\\','=']:
                    cur=current.popleft()
                    if pedantic and old_lines[0] != cur:
                        raise Error('script mismatch, expecting dialogue line %s but got %s instead' % (repr(old_lines[0]), repr(cur)))
                    old_lines = old_lines[1:]
                # linebreak the text to be inserted
                breaked_lines = []
                for input_line in _ttext.split('\\n'):
                    breaked_lines.extend(wrapper.wrap(input_line))
                # paginate first page
                output.extend(breaked_lines[0:cur_rows])
                # paginate subsequent pages
                start = cur_rows
                cur_rows -= len(breaked_lines[0:cur_rows])
                for lineno in range(start,len(breaked_lines),rows):
                    # add a pause
                    output.append('\x5c')
                    output.extend(breaked_lines[lineno:lineno+rows])
                    cur_rows = rows-len(breaked_lines[lineno:lineno+rows])
                # we're done
        while current:
            output.append(current.popleft())
        self.lines = output

    def export(self):
        """
        Return a binary string representing the script file
        """
        return '\x00'.join(self.lines) + '\x00'

def import_from_generic_file(scriptblock, filepath, options):
    """
    Import a generic MarbleEngine script file
    """
    f = open(filepath, "rb")
    s = f.read()
    f.close()
    script = MBLScript(s)

    scriptblock.clear()
    scriptblock.type = 'mbl'

    i=0
    for line in script.lines:
        sbline = scriptblock.newLine()
        sbline.otext = line.decode('cp932')
        sbline.ttext = line.decode('cp932')
        sbline.lineNum = str(i)
        sbline.speaker = 'raw'
        scriptblock.lines.append(sbline)
        scriptblock.linesTable[str(i)] = sbline
        i=i+1

    scriptblock.dirty = True
        
def import_from_binary_file(scriptblock, filepath, options):
    """
    Import a MarbleEngine script file
    """
    f = open(filepath, "rb")
    s = f.read()
    f.close()
    script = MBLScript(s)
    lines = script.dump()

    scriptblock.clear()
    scriptblock.type = 'mbl'

    speakers = {}
    i=0
    for (speaker,line) in lines:
        sbline = scriptblock.newLine()
        if speaker and speaker not in ['title','choice']:
            speakers[speaker]=True
        sbline.otext = line
        sbline.ttext = line
        sbline.lineNum = str(i)
        sbline.speaker = speaker
        scriptblock.lines.append(sbline)
        scriptblock.linesTable[str(i)] = sbline
        i=i+1

    for speaker in speakers:
        sbname = scriptblock.newName()
        sbname.oname = speaker
        sbname.tname = speaker
        scriptblock.namesTable[speaker] = sbname
        scriptblock.names.append(speaker)

    scriptblock.dirty = True

_binary_export_options = {
    'select_files':[
    ('Please choose the original binary script file',
     '','',
     'MarbleEngine script file (*)|*',False)
    ]
    }

def export_to_binary_file(scriptblock, filepath, options):
    # Set options
    if 'cols' in options:
        cols = options['cols']
    else:
        cols = 44
    if 'rows' in options:
        rows = options['rows']
    else:
        rows = 3
    # Get the original binary file loaded
    origpath = options['select_files'][0][0]
    script=MBLScript(open(origpath,"rb").read())
    # create list to insert
    new_lines = []
    for line in scriptblock.lines:
        speaker = line.speaker
        if speaker:
            # look up speaker translation
            speaker = line.speaker
            if speaker in scriptblock.namesTable:
                speaker = scriptblock.namesTable[speaker].tname
        # append
        new_lines.append((speaker,line.ttext,line.otext))
    # do insertion
    script.insert(new_lines, cols=cols, rows=rows)
    # save output
    open(filepath,"wb").write(script.export())

exports=[
    ('MarbleEngine script file','',export_to_binary_file,_binary_export_options),
    ]

imports=[
    ('MarbleEngine script file','',import_from_binary_file,{}),
    ('MarbleEngine generic script file','',import_from_generic_file,{}),
    ]

desc=('MarbleEngine extracted script file','.xml')

