"""
A module for handling CIRCUS's .mes files
"""

import struct

class Error(Exception):
    pass

DEBUG = False

def mes_tokenizer(buffer, offset, switch1=0x29, switch2=0x2e, switch3=0x4d):
    """
    Create a tokenizer that tokenizes the scripting commands
    from a .mes file

    The buffer is the actual script buffer and offset
    is where tokenization will begin
    """
    while offset < len(buffer):
        command = ord(buffer[offset])
        if command == 0:
            if offset+3 <= len(buffer):
                yield buffer[offset:offset+3]
                offset += 3
            else:
                raise Error('Unexpected end of buffer before end of token')
            break # done with file
        elif command < switch1:
            # simple 3-byte command
            if offset+3 <= len(buffer):
                yield buffer[offset:offset+3]
                offset += 3
            else:
                raise Error('Unexpected end of buffer before end of token')
        elif command < switch2:
            # 2 byte command followed by null-terminated string
            try:
                zero_location = buffer.index('\x00',offset+2)
            except ValueError:
                raise Error('Unexpected end of buffer before end of string')
            yield buffer[offset:zero_location+1]
            offset = zero_location + 1
        elif command < switch3:
            # 1 byte command followed by null-terminated string
            try:
                zero_location = buffer.index('\x00',offset+1)
            except ValueError:
                raise Error('Unexpected end of buffer before end of string')
            yield buffer[offset:zero_location+1]
            offset = zero_location + 1
        else:
            # 0x4d and up, 9 bytes total
            if offset+9 <= len(buffer):
                yield buffer[offset:offset+9]
                offset += 9
            else:
                raise Error('Unexpected end of buffer before end of token')
    if DEBUG:
        print hex(offset)
    return

_boundaries_lookup = {
    '\x68\x35':(0x29,0x2e,0x4d), # shura
    '\x66\x37':(0x2f,0x2f,0x50), # true world
}

_textopcodes_lookup = {
    '\x68\x35':'\x44\x4a\x4b\x4c', # shura
    '\x66\x37':'\x47\x4d\x4e\x4f', # true world
}

_linewidth_lookup = {
    '\x68\x35':42, # shura
    '\x66\x37':50, # true world
}

class MES:
    """
    A class tot parse .mes files
    """
    def __init__(self):
        self.clear()

    def clear(self):
        # header
        self.header = '\x68\x35'
        # parsed tokens
        self.tokens = []
        # section indices
        self.sections = []
        # unparsed script buffer
        self.buffer = ''
        # opcode boundaries
        self.boundaries = _boundaries_lookup[self.header]
        # text opcodes
        self.textopcodes = _textopcodes_lookup[self.header]
        # line width
        self.linewidth = _linewidth_lookup[self.header]

    def load(self, url):
        """
        Load a .mes file
        """
        f=open(url,'rb')
        # read the directory size
        dir_size, = struct.unpack('<L',f.read(4))
        # read the directory
        directory = struct.unpack('<'+'L'*dir_size,f.read(4*dir_size))
        # read the buffer
        buffer=f.read(100000)
        # clear the file
        self.clear()
        # get the header
        self.header = buffer[0:2]
        try:
            self.boundaries = _boundaries_lookup[self.header]
            self.textopcodes = _textopcodes_lookup[self.header]
            self.linewidth = _linewidth_lookup[self.header]
        except KeyError:
            raise Error('Unrecognized CIRCUS .mes file')
        # get the buffer
        self.buffer = buffer
        # parse the tokens (they start two bytes into the buffer)
        tokenizer = mes_tokenizer(self.buffer, 2, *self.boundaries)
        self.tokens = [x for x in tokenizer]
        # find the token offsets
        token_lengths = [len(x) for x in self.tokens]
        token_offsets = reduce(lambda x,y: x+[x[-1]+y],token_lengths,[2])
        try:
            self.sections = [token_offsets.index(x) for x in directory]
        except ValueError:
            raise Error('Could not match tokenized output to directory')

    def save(self, url):
        """
        Save a new .mes file
        """
        f=open(url,'wb')
        # coagulate the sections
        section_bins = []
        section_offsets = []
        sections_padded = self.sections + [len(self.tokens)]
        offset = 2
        cur_token = 0
        for i in range(len(sections_padded)):
            section_bin = ''.join(self.tokens[cur_token:sections_padded[i]])
            section_bins.append(section_bin)
            if i:
                section_offsets.append(offset)
            offset += len(section_bin)
            cur_token = sections_padded[i]
        # write the file
        f.write(struct.pack('<L',len(section_offsets))) # number of section offsets
        f.write(struct.pack('<'+'L'*len(section_offsets),*section_offsets)) # actual offsets
        f.write('\x68\x35') # Circus magic number
        f.write(''.join(section_bins)) # all the sections
        f.close()
            
