'''
A module for loading .MSD (MSCENARIO) files from Fuguriya game(s)
'''

import md5
import struct
import StringIO

class Error(Exception):
    pass

# the seeds needed for decryption
_key_seeds = {
    'sonohana':'\x82\xbb\x82\xcc\x89\xd4\x82\xd1\x82\xe7\x82\xc9\x82\xad\x82\xbf\x82\xc3\x82\xaf\x82\xf0',
    }

def _md5_digest(s):
    '''
    Return a plaintext MD5 digest of a string
    '''
    return md5.new(s).hexdigest()

def _crypt(s, key_type):
    '''
    Apply the crypto function used for encrypting script files
    '''
    key_seed = _key_seeds[key_type]
    output = []
    for i in range((len(s)+31)/32):
        in_block = s[i*32:i*32+32]
        key = _md5_digest(key_seed+('%d' % i))
        key = key[0:len(in_block)]
        out_block = ''.join(map(lambda x,y: chr(ord(x) ^ ord(y)),in_block,key))
        output.append(out_block)
    return ''.join(output)

def _read_string(s):
    i = 0
    state = 0
    for ch in s:
        i += 1
        if state == 1:
            if ch == '&': # an escaped ampersand
                state = 0
            else:
                state = 2
        elif state == 2:
            continue
        if ch == '\x00':
            return s[0:i]
        if ch == '&':
            state = 1
    raise Error('Unexpected end of string')

def _parse_params(param_block):
    '''
    Tokenize the parameters
    '''
    index = 0
    parameters = []
    # print 'parse params'
    while index < len(param_block):
        # print index
        # read a flag character for type
        param_type = ord(param_block[index])
        if param_type == 1:
            # 32bit parameter
            parameter = param_block[index:index+5]
            if len(parameter) != 5:
                raise Error('Unexpected end of statement')
            parameters.append(parameter)
            index += 5
        elif param_type == 2:
            # 32bit parameter
            parameter = param_block[index:index+5]
            if len(parameter) != 5:
                raise Error('Unexpected end of statement')
            parameters.append(parameter)
            index += 5
        elif param_type == 3:
            # string parameter
            s = _read_string(param_block[index+1:])
            parameters.append('\x03'+s)
            index += len(s)+1
        elif param_type == 4:
            # tuple parameter
            index += 1
            entries = []
            entry = param_block[index:index+8]
            index += 8
            if len(entry) != 8:
                raise Error('Unexpected end of statement')
            entries.append(entry)
            while entry[0:4] != '\xff\xff\xff\xff':
                entry = param_block[index:index+8]
                index += 8
                if len(entry) != 8:
                    raise Error('Unexpected end of statement')
                entries.append(entry)
            parameter = '\x04'+''.join(entries)
            parameters.append(parameter)
        elif param_type == 5:
            # dictionary entry parameter
            # 32bit parameter
            parameter = param_block[index:index+5]
            if len(parameter) != 5:
                raise Error('Unexpected end of statement')
            parameters.append(parameter)
            index += 5
        else:
            print repr(parameters)
            raise Error('Unknown parameter type: %x' % param_type)
    return parameters

def _read_code_line(f):
    '''
    Read in a line of machine code from script file
    '''
    opcode = f.read(2)
    if len(opcode) == 0:
        return
    elif len(opcode) == 1:
        raise Error('Unexpected end of file when reading statement')
    # print 'Opcode:',opcode.encode('hex')
    param_length_str = f.read(2)
    if len(param_length_str) != 2:
        raise Error('Unexpected end of file when reading statement')
    param_length, = struct.unpack('<H',param_length_str)
    # print 'Param length:',hex(param_length)
    # read the parameter block
    param_block = f.read(param_length)
    # print ' '.join([x.encode('hex') for x in [opcode,param_length_str]+_parse_params(param_block)])
    return [opcode,param_length_str]+_parse_params(param_block)

def decrypt_file(url_in, url_out, key_type='sonohana'):
    f = open(url_in,'rb')
    data = f.read()
    f.close()
    data = _crypt(data,key_type)
    f = open(url_out,'wb')
    f.write(data)
    f.close()

class MSD:
    def __init__(self, key_type='sonohana'):
        self.clear()
        self.key_type = 'sonohana'

    def clear(self):
        self.version = 1,0
        self.labels = [] # hex offsets for jump labels
        self.load_points = [] # hex offsets for load points
        self.code = [] # a list of token lists

    def load(self, url):
        f=open(url,'rb')
        if f.read(0x10) != 'MSCENARIO\x20FILE\x20\x20':
            # try decrypting it
            f.seek(0)
            s=f.read()
            s_decrypted = _crypt(s,self.key_type)
            if s_decrypted[0:0x10] != 'MSCENARIO\x20FILE\x20\x20':
                raise Error('Not a MSCENARIO file')
            f.close()
            f=StringIO.StringIO(s_decrypted)
            f.seek(0x10)
        versionminor,versionmajor = struct.unpack('<HH',f.read(4))
        self.version = versionmajor,versionminor
        if self.version != (1,0):
            raise Error('Expecting a version 1.0 MSCENARIO file, got %d.%d instead' % self.version)
        self.clear()
        # read number of jump labels and number of load points
        num_labels, num_load_points = struct.unpack('<LL',f.read(8))
        # read labels
        f.seek(0x458)
        label_str = f.read(4*num_labels)
        self.labels = list(struct.unpack('<'+'L'*num_labels,label_str))
        # read load points
        load_point_str = f.read(4*num_load_points)
        self.load_points = list(struct.unpack('<'+'L'*num_load_points,load_point_str))
        # read the actual code
        # print hex(f.tell())
        code_line = _read_code_line(f)
        while code_line:
            # print hex(f.tell())
            self.code.append(code_line)
            code_line = _read_code_line(f)
        f.close()

    def get_strings(self):
        '''
        Get a list of Japanese strings in unicode format
        '''
        string_list = []
        for code_line in self.code:
            if code_line[0] == '\xda\x07':
                # normal string
                string_arg = code_line[5]
            elif code_line[0] == '\xed\x03':
                # chapter title
                string_arg = code_line[2]
            else:
                string_arg = None
            if string_arg:
                if string_arg[0] != '\x03':
                    raise Error('String argument expected, got type %d instead' % ord(string_arg[0]))
                unicode_string = string_arg[1:-1].decode('shift-jisx0213')
                string_list.append(unicode_string)
        return string_list

    def get_dump(self):
        '''
        Get a dumped list of code
        '''
        string_list = []
        for code_line in self.code:
            start = 'func'+code_line[0].encode('hex')
            arg_list = []
            for arg in code_line[2:]:
                if arg[0] == '\x01':
                    arg_list.append(str(struct.unpack('<l',arg[1:])[0]))
                elif arg[0] == '\x02':
                    arg_list.append('0'+str(struct.unpack('<L',arg[1:])[0]))
                elif arg[0] == '\x03':
                    arg_list.append(arg[1:-1].decode('shift-jisx0213'))
                elif arg[0] == '\x04':
                    arg_list.append('tuple')
                elif arg[0] == '\x05':
                    arg_list.append('@'+str(struct.unpack('<L',arg[1:])[0]))
            string_list.append(start+'('+(','.join(arg_list))+')')
        return string_list

    def replace(self, new_code):
        '''
        Replace the current code with new code
        '''
        if len(self.code) != len(new_code):
            raise Error('Replacement code must have the same number of lines')
        old_lengths = [len(''.join(x)) for x in self.code]
        new_lengths = [len(''.join(x)) for x in new_code]
        old_offsets = reduce(lambda x,y: x + [x[-1]+y], old_lengths, [0])
        new_offsets = reduce(lambda x,y: x + [x[-1]+y], new_lengths, [0])
        offset_pairs = zip(old_offsets,new_offsets)
        offset_lookup = dict(offset_pairs)
        new_labels = [offset_lookup[x] for x in self.labels]
        new_load_points = [offset_lookup[x] for x in self.load_points]
        self.code = new_code
        self.labels = new_labels
        self.load_points = new_load_points

    def save(self, url):
        f=open(url,'wb')
        # write header
        f.write('MSCENARIO\x20FILE\x20\x20')
        # write version?
        f.write('\x00\x00\x01\x00')
        # write number of jump labels and number of load points
        f.write(struct.pack('<LL',len(self.labels),len(self.load_points)))
        # fill rest with 0's
        f.write('\x00'*(0x458-f.tell()))
        # write the labels
        f.write(struct.pack('<'+'L'*len(self.labels),*self.labels))
        # write the load points
        f.write(struct.pack('<'+'L'*len(self.load_points),*self.load_points))
        # write the code
        f.write(''.join([''.join(x) for x in self.code]))
        
