'''
mscenario script plugin
'''

import mscenario
import textwrap
import struct

def decode(s):
    '''
    Decode game script string by making control codes human readable
    '''
    # convert out of shiftjis
    s=s.decode('shift-jisx0213')
    return s

def encode(s):
    '''
    Encode game script string
    '''
    # convert into shiftjis
    s=s.encode('shift-jisx0213')
    return s

_basic_table = ''.join([chr(x) for x in range(256)])

_speaker_table = {
    '\x01\x00\x00\x00\x00':'Nanami',
    '\x01\x01\x00\x00\x00':'Yuuna',
    '\x01\x02\x00\x00\x00':'Girl A',
    '\x01\x03\x00\x00\x00':'Girl B',
    '\x01\x04\x00\x00\x00':'Girl C',
    '\x01\x05\x00\x00\x00':'Teacher',
    '\x01\x06\x00\x00\x00':'Kid A',
    '\x01\x07\x00\x00\x00':'Kid B',
    '\x01\x08\x00\x00\x00':'???',
    '\x01\xff\xff\xff\xff':'',
    }

def import_from_MSD_file(scriptblock, filepath, options):
    '''
    Import the script from a .SNX file
    '''
    x=mscenario.MSD()
    x.load(filepath)
    scriptblock.clear()
    scriptblock.type = 'msd'
    scriptblock.dirty = True
    i = 0
    cur_speaker = ''
    for code_line in x.code:
        if code_line[0] == '\xd7\x07':
            # change speaker
            cur_speaker = _speaker_table[code_line[2]]
        elif code_line[0] == '\xda\x07':
            # normal text
            line = decode(code_line[5][1:-1])
            sbline = scriptblock.newLine()
            sbline.otext = line
            sbline.ttext = line
            sbline.lineNum = str(i)
            sbline.speaker = cur_speaker
            scriptblock.lines.append(sbline)
            scriptblock.linesTable[str(i)] = sbline
            i=i+1
        elif code_line[0] == '\xed\x03':
            # chapter title
            line = decode(code_line[2][1:-1])
            sbline = scriptblock.newLine()
            sbline.otext = line
            sbline.ttext = line
            sbline.lineNum = str(i)
            sbline.speaker = 'Title'
            scriptblock.lines.append(sbline)
            scriptblock.linesTable[str(i)] = sbline
            i=i+1

def export_to_text_file(scriptblock, filepath, options):
    '''
    Export the script to a text file
    '''
    # open the file
    f = open(filepath,'wb')
    for line in scriptblock.lines:
        if line.lineNum >= 0:
            textline = ''
            speaker = line.speaker
            if speaker in scriptblock.namesTable:
                speaker = scriptblock.namesTable[speaker].tname
            if speaker:
                textline+=(speaker+': ')
            textline+=(line.ttext)
            if line.comment.strip():
                textline+=(' //'+line.comment.strip())
            f.write(textline+'\n')
        else:
            f.write('\n')
    f.close()

_MSD_export_options = {
    'select_files':[
    ('Please choose the original MSD file',
     '','',
     'mscenario script file (*.MSD)|*.MSD',False)
    ]
    }

def _pad(s, columns):
    additional = columns-len(s)
    double_spaces = additional/2
    single_spaces = additional%2
    # s = s+'\x81\x40'*double_spaces# + ' '*single_spaces
    s = s+' '*additional
    return s

def export_to_MSD_file(scriptblock, filepath, options):
    # Get the original MSD file loaded
    origpath = options['select_files'][0][0]
    x=mscenario.MSD()
    x.load(origpath)
    # Get the word wrap options
    if 'columns' in options:
        columns = options['columns']
    else:
        columns = 44
    if 'rows' in options:
        rows = options['rows']
    else:
        rows = 3
    # Create the linewrapper
    wrapper = textwrap.TextWrapper(width=columns,break_long_words=False)
    # Reformat the script
    strings = []
    for line in scriptblock.lines:
        newline = encode(line.ttext)
        # do linewrapping
        DO_WRAP = True
        if DO_WRAP:
            #newline = '  '+newline
            wrapped_lines = wrapper.wrap(newline)
            wrapped_lines = ['\x81\x40'+_pad(y,columns)+'\x81\x40' for y in wrapped_lines]
            newline = ''.join(wrapped_lines)
            newline = newline[2:-2]
            newline = newline.rstrip(' ')
            #newline = newline[2:]
        # append
        strings.append('\x03'+newline+'\x00')
    # Create a new set of codes
    new_code = []
    index = 0
    for code_line in x.code:
        new_code_line = code_line[:]
        if code_line[0] == '\xda\x07':
            new_code_line[5] = strings[index]
            index += 1
        elif code_line[0] == '\xed\x03':
            new_code_line[2] = strings[index]
            index += 1
        new_code_line[1] = struct.pack('<H',len(''.join(new_code_line[2:])))
        new_code.append(new_code_line)
    # Insert the script
    x.replace(new_code)
    # Save to the new MSD file
    x.save(filepath)

# the following must be present in each plugin file

exports=[
    ('MSCENARIO script file','.MSD',export_to_MSD_file,_MSD_export_options),
    ('Text file','.txt',export_to_text_file,{}),
    ]

imports=[
    ('MSCENARIO script file','.MSD',import_from_MSD_file,{}),
    ]

desc=('MSCENARIO xml script file','.msx')
