# A module for handling Nexton's SNX script files used in ONE FV
import struct
import os

class Error(Exception):
    pass

class SNX:
    def __init__(self, path=None, pedantic=False):
        if path:
            self.load(path, pedantic)
        else:
            self.clear()

    def clear(self):
        self.code = []
        self.text = []

    def load(self, path, pedantic=False):
        '''
        Load a script file
        '''
        self.clear()
        f=open(path,'rb')
        header = f.read(8)
        code_length, text_length = struct.unpack('<LL',header)
        for i in range(code_length):
            # read the code
            self.code.append(f.read(0xc))
        if pedantic:
            end_position = f.tell()+text_length
        else:
            end_position = os.path.getsize(path)
        while f.tell() < end_position:
            # read the text
            string_length, = struct.unpack('<L',f.read(4))
            cur_string = f.read(string_length)
            if cur_string[-1] != '\x00':
                raise Error('Error while reading string number %d: Could not find end of string.' % (i+1))
            self.text.append(cur_string[0:-1])
        f.close()

    def replace(self, new_text):
        '''
        Replace the old set of text with new set of text, and rewrite pointers
        '''
        # build up a list of text offsets and create a mapping
        old_offsets = reduce(lambda x,y: x+[x[-1]+len(y)+5],self.text,[0])
        new_offsets = reduce(lambda x,y: x+[x[-1]+len(y)+5],new_text,[0])
        offset_pairs = map(lambda x,y: [x,y],old_offsets,new_offsets)
        offset_map = dict(offset_pairs)
        new_code = []
        for instruction in self.code:
            if instruction[0:8] == '\x11\x00\x00\x00\x02\x00\x00\x00':
                pointer, = struct.unpack('<L',instruction[8:0xc])
                if pointer not in offset_map:
                    raise Error('Illegal text reference instruction encountered: '+hex(pointer))
                instruction = '\x11\x00\x00\x00\x02\x00\x00\x00'+struct.pack('<L',offset_map[pointer])
            new_code.append(instruction)
        # set the new code and text
        self.code = new_code
        self.text = new_text

    def save(self, path):
        '''
        Save a script file
        '''
        code_length = len(self.code)
        code_data = ''.join(self.code)
        if len(code_data) != code_length * 0xc:
            raise Error('Code not of current length, should be %d, but got %d instead' % (code_length*0xc,len(code_data)))
        text_length = len(self.text)*5 + len(''.join(self.text))
        header = struct.pack('<LL',code_length,text_length)
        f=open(path,'wb')
        f.write(header)
        f.write(code_data)
        for cur_string in self.text:
            f.write(struct.pack('<L',len(cur_string)+1))
            f.write(cur_string+'\x00')
        f.close()
                        
