'''
ONE Full Voice version plugin
'''

from nexton import *
import textwrap
import struct
import re

_byte2_flag = [chr(x) for x in range(0x81,0xa0) + range(0xe0,0xfd)]

def decode(s):
    '''
    A more complete decoder for in game text
    '''
    consume = 0
    discard = 0
    output = []
    slen = len(s)
    i = 0
    while i < slen:
        ch = s[i]
        if ch in _byte2_flag:
            # 2 byte shift jis section
            output.append(s[i:i+2])
            i+=2
        elif ch == '\x01':
            # newline
            output.append('\\n')
            i+=1
        elif ch == '\x02':
            # pause
            output.append('\\p')
            i+=1
        elif ch == '\x03':
            # clear
            output.append('\\c')
            i+=1
        elif ch == '\x04':
            # number entry
            entry_num, = struct.unpack('<L',s[i+1:i+5])
            output.append('\\i('+str(entry_num)+')')
            i+=5
        elif ch == '\x05':
            # dictionary entry
            entry_num, = struct.unpack('<L',s[i+1:i+5])
            output.append('\\d('+str(entry_num)+')')
            i+=5
        elif ch == '\x5c':
            # backslash escape
            output.append('\\\\')
            i+=1
        else:
            output.append(ch)
            i+=1
    try:
        return ''.join(output).decode('cp932')
    except Exception, e:
        print ''.join(output).encode('hex')
        print ''.join(output)
        raise e

_dict_pattern = r'\\d(\([0-9]+\))'
_number_pattern = r'\\i(\([0-9]+\))'

def _dict_replace(match_object):
    return '\x05' + struct.pack('<L',int(match_object.group()[3:-1]))

def _number_replace(match_object):
    return '\x04' + struct.pack('<L',int(match_object.group()[3:-1]))

_dict_searcher = re.compile(_dict_pattern)
_number_searcher = re.compile(_number_pattern)

_single_lookup = {'n':'\x01',
                  'p':'\x02',
                  'c':'\x03',
                  '\\':'\\'}

_parenlong_lookup = {'i':'\x04',
                     'd':'\x05'}

def _word_type(s):
    ch = s[0]
    if ch in _byte2_flag:
        return 'double' # double wide character, can break anywhere
    elif ord(ch) in range(0x21,0x7f):
        return 'single' # single wide character, keep as a word
    elif ord(ch) in range(0x01,0x05):
        return 'control' # control code
    elif ch == '\x20':
        return 'space' # whitespace, etc.
    else:
        raise Error('Unexpected character: %02X' % ord(ch))

def _word_space(s, dict_length=8, number_length=8):
    ch = s[0]
    if ch in _byte2_flag:
        if len(s) != 2:
            raise Error('Illegal double byte sequence: %s' % s.encode('hex'))
        return 2
    elif ch == '\x01':
        # clear the line
        return -1
    elif ch == '\x03':
        # clear the box
        return -2
    elif ch == '\x04':
        # dictionary
        return dict_length
    elif ch == '\x05':
        # number lookup
        return number_length
    elif ch == '\x20':
        return 1 # a single space
    elif ord(ch) in range(0x21,0x7f):
        return len(s) # single byte character word
    else:
        return 0 # no spacing needed

def encode(s, orig, columns=42, initial_indent=1,
           subsequent_indent=1, right_margin=2, rows=4,
           dict_length=8, number_length=8):
    '''
    A more complete game script encoder
    Will auto line break as well by using \x01 (\n)
    '''
    def indent_string(current_row):
        if current_row:
            return ' '*subsequent_indent
        else:
            return ' '*initial_indent

    orig_equal = (s == orig)
    s = s.encode('cp932')
    slen = len(s)
    i = 0
    output = []
    do_linebreak = False
    while i < slen:
        ch = s[i]
        if ch in _byte2_flag:
            # 2 byte shift jis section
            output.append(s[i:i+2])
            i+=2
        elif ch == '\x5c':
            # backslash
            i+=2
            ch2 = s[i-1]
            if ch2 in ['n','p','c']:
                do_linebreak = True
            if ch2 in _single_lookup:
                output.append(_single_lookup[ch2])
            elif ch2 in _parenlong_lookup:
                output.append(_parenlong_lookup[ch2])
                # find the parenthesized expression
                if s[i] != '(':
                    raise Error('Expecting ( after a \\i or a \\d')
                end_paren = s.find(')',i)
                if end_paren < i:
                    raise Error('Could not find ending ) after a \\id or a \\d')
                try:
                    value = int(eval(s[i+1:end_paren]))
                except Exception:
                    raise Error('Unable to evaluate expression %s' % s[i+1:end_paren])
                output.append(struct.pack('<L',value))
                i = end_paren+1
            else:
                raise Error("Unexpected escaped character '%s'" % ch2)
        else:
            i+=1
            output.append(ch)
    # if we don't need to linebreak or if original is the same, return here
    if not do_linebreak or orig_equal:
        return ''.join(output)
    # calculate geometry
    line_width_1 = columns - initial_indent - right_margin
    line_width_2 = columns - subsequent_indent - right_margin
    min_line_width = min(line_width_1,line_width_2)
    # coagulate tokens into words
    current_word = ''
    words = []
    for token in output:
        if _word_type(token) == 'single':
            # append to word
            current_word += token
        else:
            # write out current word, if there's any
            if current_word:
                if len(current_word) > min_line_width:
                    words.extend([current_word[i:i+min_line_width] for i in range(0,len(current_word),min_line_width)])
                else:
                    words.append(current_word)
            if len(token) > min_line_width:
                words.extend([token[i:i+min_line_width] for i in range(0,len(token),min_line_width)])
            else:
                words.append(token)
            current_word = ''
    # break into lines
    current_row = 0
    current_line = indent_string(current_row)
    current_len = 0
    lines = []
    for word in words:
        if current_row == 0:
            line_width = line_width_1
        else:
            line_width = line_width_2
        space = _word_space(word, dict_length, number_length)
        if space == -1:
            # explicit new line
            if current_row >= rows-1:
                new_line = current_line+'\x02\x03'
                current_row = 0
            else:
                new_line = current_line+word
                current_row += 1
            lines.append(new_line)
            current_line = indent_string(current_row)
            current_len = 0
        elif space == -2:
            # explicit new paragraph
            lines.append(current_line+word)
            current_row = 0
            current_line = indent_string(current_row)
            current_len = 0
        elif space + current_len > line_width:
            # implicit new line
            # this line is full, create a new one
            if current_row >= rows-1:
                new_line = current_line+'\x02\x03'
                current_row = 0
            else:
                new_line = current_line+'\x01'
                current_row += 1
            lines.append(new_line)
            current_line = indent_string(current_row)+word
            current_len = space
        else:
            current_line += word
            current_len += space
    return ''.join(lines)

class NotASpeaker(Exception):
    pass

def process_speaker(s):
    '''
    Take a decoded string, and try to parse out the speaker
    '''
    # find the left quote symbol
    left_location = s.find(u'\u300c')
    # find the right quote symbol
    right_location = s.rfind(u'\u300d')
    # throw an exception if not a speaker
    if left_location < 0 or right_location < 0 or left_location > right_location:
        raise NotASpeaker()
    # speaker is to the left
    speaker = s[0:left_location]
    # the rest is text
    line = s[left_location:]
    if speaker == '':
        speaker = '(none)'
    return speaker, line

def import_from_SNX_file(scriptblock, filepath, options):
    '''
    Import the script from a .SNX file
    '''
    x=SNX()
    x.load(filepath)
    scriptblock.clear()
    scriptblock.type = 'onefv'
    i = 0
    speakers = {}
    for text_line in x.text:
        decoded_line = decode(text_line)
        try:
            speaker, line = process_speaker(decoded_line)
            speakers[speaker] = True
        except NotASpeaker:
            speaker = ''
            line = decoded_line
        sbline = scriptblock.newLine()
        sbline.otext = line
        sbline.ttext = line
        sbline.lineNum = str(i)
        sbline.speaker = speaker
        scriptblock.lines.append(sbline)
        scriptblock.linesTable[str(i)] = sbline
        i=i+1

    for speaker in speakers:
        sbname = scriptblock.newName()
        sbname.oname = speaker
        sbname.tname = speaker
        scriptblock.namesTable[speaker] = sbname
        scriptblock.names.append(speaker)

    scriptblock.dirty = True

def export_to_text_file(scriptblock, filepath, options):
    '''
    Export the script to a text file
    '''
    # open the file
    f = open(filepath,'wb')
    for line in scriptblock.lines:
        if line.lineNum >= 0:
            textline = ''
            speaker = line.speaker
            if speaker in scriptblock.namesTable:
                speaker = scriptblock.namesTable[speaker].tname
            if speaker:
                textline+=(speaker+': ')
            textline+=(line.ttext)
            if line.comment.strip():
                textline+=(' //'+line.comment.strip())
            f.write(textline+'\n')
        else:
            f.write('\n')
    f.close()

_SNX_export_options = {
    'select_files':[
    ('Please choose the original SNX file',
     '','',
     'NEXTON script file (*.SNX)|*.SNX',False)
    ]
    }

def export_to_SNX_file(scriptblock, filepath, options):
    # Get the original SNX file loaded
    origpath = options['select_files'][0][0]
    x=SNX()
    x.load(origpath)
    # Get the word wrap options
    if 'columns' in options:
        columns = options['columns']
    else:
        columns = 42
    if 'initial_indent' in options:
        initial_indent = options['initial_indent']
    else:
        initial_indent = 1
    if 'subsequent_indent' in options:
        subsequent_indent = options['subsequent_indent']
    else:
        subsequent_indent = 1
    if 'right_margin' in options:
        right_margin = options['right_margin']
    else:
        right_margin = 2
    if 'rows' in options:
        rows = options['rows']
    else:
        rows = 4
    if 'dict_length' in options:
        dict_length = options['dict_length']
    else:
        dict_length = 8
    if 'number_length' in options:
        number_length = options['number_length']
    else:
        number_length = 8
    # Reformat the script
    strings = []
    for line in scriptblock.lines:
        if line.speaker:
            # dialog, need to put together
            speaker = line.speaker
            if speaker in scriptblock.namesTable:
                speaker = scriptblock.namesTable[speaker].tname
            if speaker == '(none)':
                speaker = ''
            else:
                speaker += ': '
            newline = speaker+line.ttext
            oldline = speaker+line.otext
        else:
            speaker = ''
            newline = line.ttext
            oldline = line.otext
        newline = encode(newline, oldline,
                         columns=columns,
                         initial_indent=initial_indent,
                         subsequent_indent=subsequent_indent,
                         right_margin=right_margin,
                         rows=rows,
                         dict_length=dict_length,
                         number_length=number_length)
        strings.append(newline)
    # Insert the script
    x.replace(strings)
    # Save to the new SNX file
    x.save(filepath)

# the following must be present in each plugin file

exports=[
    ('NEXTON script file','.SNX',export_to_SNX_file,_SNX_export_options),
    ('Text file','.txt',export_to_text_file,{}),
    ]

imports=[
    ('NEXTON script file','.SNX',import_from_SNX_file,{}),
    ]

desc=('NEXTON xml script file','.nxs')
