import mmap
import os.path
import struct

DEBUG = False

class Error(Exception):
    pass

class RVATable:
    '''
    A class to handle virtual addressing
    '''
    def __init__(self, old_table=None):
        if not old_table:
            self.blocks = [] # a sorted list of sections in ascending order of starting address
            self.dirty = False # if blocks need sorting
        else:
            # make a copy
            self.blocks = old_table.blocks[:]
            self.dirty = old_table.dirty

    def sort_blocks(self):
        '''
        Sort the blocks in ascending order of starting address
        '''
        self.blocks.sort(lambda x,y: cmp(x[1],y[1]))
        self.dirty = False

    def check_collisions(self):
        '''
        Check for collisions between the blocks
        '''
        if not self.blocks:
            return
        if self.dirty:
            self.sort_blocks()
        last_end = self.blocks[0][1]
        last_name = ''
        for name,start,length in self.blocks:
            if start < last_end:
                raise Error("A collision has occurred between blocks '"+last_name+"' and '"+name+"'")
            last_end = start + length
            last_name = name

    def add_block(self, name, start, length):
        '''
        Add a block
        '''
        self.blocks.append((name,start,length))
        self.dirty = True

    def lookup(self, value):
        '''
        Look up a value to find which block it resides in and the offset in that block
        Returns (blockname, offset)
        If no possible block, return (None, value)
        If there are collisions, the value is undefined
        '''
        if self.dirty:
            self.sort_blocks()
        poss_blocks = [x for x in self.blocks if x[1] <= value]
        if not poss_blocks:
            return (None, value)
        block = poss_blocks[-1]
        offset = value - block[1]
        if offset >= block[2]:
            return (None, value)
        return block[0], offset

    def len(self):
        '''
        Return the highest value spanned by this table
        '''
        if not self.blocks:
            return 0
        if self.dirty:
            self.sort_blocks()
        return self.blocks[-1][1]+self.blocks[-1][2]
        
class MemoryView:
    '''
    A class for accessing a PESections object like a string
    '''
    def __init__(self, pe_sections):
        self.pe_sections = pe_sections

    def __len__(self):
        return self.pe_sections.address_table.len()

    def _get_slice(self, key):
        if type(key) == type(slice(1,1,1)):
            slice_start = key.start
            slice_end = key.stop
            if key.step:
                raise Error('Steps are not allowed in slices of MemoryView objects')
        else:
            slice_start = int(key)
            slice_end = slice_start + 1
        if slice_start < 0:
            slice_start += self.pe_sections.address_table.len()
        if slice_end < 0:
            slice_end += self.pe_sections.address_table.len()
        name_start,offset_start = self.pe_sections.address_table.lookup(slice_start)
        if None == name_start:
            raise Error('Start of memory slice is not in any section')
        name_end,offset_end = self.pe_sections.address_table.lookup(slice_end-1)
        if name_end != name_start:
            raise Error('Cannot slice across sections')
        offset_end = offset_end + 1
        return name_start,offset_start,offset_end

    def __getitem__(self, key):
        name,start,end = self._get_slice(key)
        if name not in self.pe_sections.section_lookup:
            raise Error('Fatal error, section '+name+' not found in pefile')
        section = self.pe_sections.sections[self.pe_sections.section_lookup[name]]
        return section['Data'][start:end]

    def __setitem__(self, key, value):
        name,start,end = self._get_slice(key)
        if name not in self.pe_sections.section_lookup:
            raise Error('Fatal error, section '+name+' not found in pefile')
        section = self.pe_sections.sections[self.pe_sections.section_lookup[name]]
        try:
            section['Data'][start:end] = value
        except Exception:
            new_section = section['Data'][0:start] + value[0:end-start] + section['Data'][end:]
            section['Data'] = new_section

class PESections:
    '''
    A class to handle sections in a PE file
    '''
    def __init__(self, path, extra=False):
        '''
        By default, initialize a PESections object from a pre-existing PE file
        Set extra to True if you want to load up appended data as well
        '''
        self.load(path, extra)

    def load(self, path, extra):
        '''
        Load the sections from a PE file
        '''
        # First, get a memory mapped view of the file
        filesize = os.path.getsize(path)
        fp = open(path, 'rb')
        exe_data = mmap.mmap(fp.fileno(), filesize, access=mmap.ACCESS_READ)
        fp.close()

        # Now process the DOS specific stuff
        if exe_data[0:2] != 'MZ':
            raise Error('Not a valid DOS EXE file')
        pe_header_offset, = struct.unpack('<H',exe_data[0x3c:0x3e])
        self.dos_block = exe_data[0:pe_header_offset]

        # Grab the IMAGE_FILE_HEADER
        self.pe_header_data = exe_data[pe_header_offset:pe_header_offset+0x18]
        self.pe_header = {}
        (self.pe_header['Magic'],
         self.pe_header['Machine'],
         self.pe_header['NumberOfSections'],
         self.pe_header['TimeDateStamp'],
         self.pe_header['PointerToSymbolTable'],
         self.pe_header['NumberOfSymbols'],
         self.pe_header['SizeOfOptionalHeader'],
         self.pe_header['Characteristics'],
         ) = struct.unpack('<4sHHLLLHH',self.pe_header_data)
        if self.pe_header['Magic'] != 'PE\x00\x00':
            raise Error('Not a valid PE file')

        # Grab the IMAGE_OPTIONAL_HEADER
        opt_header_size = self.pe_header['SizeOfOptionalHeader']
        self.opt_header_data = exe_data[pe_header_offset+0x18:pe_header_offset+0x18+opt_header_size]
        self.opt_header = {}
        (self.opt_header['Magic'],
         self.opt_header['MajorLinkerVersion'],
         self.opt_header['MinorLinkerVersion'],
         self.opt_header['SizeOfCode'],
         self.opt_header['SizeOfInitializedData'],
         self.opt_header['SizeOfUninitializedData'],
         self.opt_header['AddressOfEntryPoint'],
         self.opt_header['BaseOfCode'],
         self.opt_header['BaseOfData'],
         self.opt_header['ImageBase'],
         self.opt_header['SectionAlignment'],
         self.opt_header['FileAlignment'],
         self.opt_header['MajorOperatingSystemVersion'],
         self.opt_header['MinorOperatingSystemVersion'],
         self.opt_header['MajorImageVersion'],
         self.opt_header['MinorImageVersion'],
         self.opt_header['MajorSubsystemVersion'],
         self.opt_header['MinorSubsystemVersion'],
         self.opt_header['Reserved1'],
         self.opt_header['SizeOfImage'],
         self.opt_header['SizeOfHeaders'],
         self.opt_header['CheckSum'],
         self.opt_header['Subsystem'],
         self.opt_header['DllCharacteristics'],
         self.opt_header['SizeOfStackReserve'],
         self.opt_header['SizeOfStackCommit'],
         self.opt_header['SizeOfHeapReserve'],
         self.opt_header['SizeOfHeapCommit'],
         self.opt_header['LoaderFlags'],
         self.opt_header['NumberOfRvaAndSizes'],
         ) = struct.unpack('<HBBLLLLLLLLLHHHHHHLLLLHHLLLLLL',self.opt_header_data[0:0x60])
        data_directories = []
        for i in range(self.opt_header['NumberOfRvaAndSizes']):
            data_directories.append(struct.unpack('<LL',self.opt_header_data[0x60+8*i:0x68+8*i]))
        self.opt_header['DataDirectory'] = data_directories
                                    
        # Grab the section table
        section_table_size = self.pe_header['NumberOfSections']*40
        section_table_start = pe_header_offset+0x18+opt_header_size
        section_table_end = section_table_start + section_table_size
        self.section_table_data = exe_data[section_table_start:section_table_end]
        self.sections = []
        self.section_lookup = {} # gets the index for self.sections if we want a particular section

        last_section_end = 0
        for i in range(self.pe_header['NumberOfSections']):
            section = {}
            (section['Name'],
             # PhysicalAddress is actually unioned with VirtualSize
             section['VirtualSize'],
             section['VirtualAddress'],
             section['SizeOfRawData'],
             section['PointerToRawData'],
             section['PointerToRelocations'],
             section['PointerToLinenumbers'],
             section['NumberOfRelocations'],
             section['NumberOfLinenumbers'],
             section['Characteristics'],
             ) = struct.unpack('<8sLLLLLLHHL',self.section_table_data[i*40:(i+1)*40])
            section['Name'] = section['Name'].rstrip('\x00')
            section['PhysicalAddress'] = section['VirtualSize']
            # get the actual data for the section
            section_start = section['PointerToRawData']
            section_end = section_start + section['SizeOfRawData']
            section['Data'] = exe_data[section_start:section_end]
            last_section_end = max(section_end,last_section_end)
            self.section_lookup[section['Name']] = i
            self.sections.append(section)

        # make the RVA table
        self.address_table = RVATable()
        for section in self.sections:
            self.address_table.add_block(section['Name'],section['VirtualAddress']+self.opt_header['ImageBase'],section['VirtualSize'])
        self.address_table.check_collisions()

        # if we need to load up extra data
        if extra:
            self.extra_data = exe_data[last_section_end:]
        else:
            self.extra_data = ''

        # close the memory mmap
        exe_data.close()

    def get_view(self):
        '''
        Return a MemoryView of the PE file
        This allows easy access to the PE file as if it were loaded into virtual memory
        '''
        return MemoryView(self)

    def add_section(self, name, virtual_address, virtual_size, characteristics, data):
        '''
        Add a new section
        '''
        if len(name) > 8:
            raise Error('Section name too long: '+name)
        if name in self.section_lookup:
            raise Error('Section already exists: '+name)
        try:
            # check for collisions
            new_rva = RVATable(self.address_table)
            new_rva.add_block(name, virtual_address+self.opt_header['ImageBase'], virtual_size)
            self.address_table = new_rva
        except Error:
            raise Error('Another section overlaps this section: '+hex(virtual_address)+'->'+hex(virtual_address+virtual_size))
        section = {}
        section['Name'] = name
        # PhysicalAddress is actually unioned with VirtualSize
        section['VirtualSize'] = virtual_size
        section['VirtualAddress'] = virtual_address
        section['SizeOfRawData'] = len(data)
        section['PointerToRawData'] = 0 # will be fixed when saving
        section['PointerToRelocations'] = 0
        section['PointerToLinenumbers'] = 0
        section['NumberOfRelocations'] = 0
        section['NumberOfLinenumbers'] = 0
        section['Characteristics'] = characteristics
        section['Data'] = data
        section['PhysicalAddress'] = section['VirtualSize']
        self.section_lookup[section['Name']] = len(self.sections)
        self.sections.append(section)

    def del_section(self, name):
        '''
        Delete a section
        '''
        index=self.section_lookup[name]
        del self.section_lookup[name]
        del self.sections[index]

    def save(self, path):
        '''
        Write out the data onto a new exe file
        Note that it will write out invalid PE files if you have been
        tampering with the header fields
        '''
        # compile the section table
        section_table = ''
        file_offset = 0x1000 # start at a safe distance
        if file_offset % self.opt_header['FileAlignment']:
            # pad up the file_offset if necessary to keep alignment
            file_offset += ((-file_offset) % self.opt_header['FileAlignment'])
        size_of_headers = file_offset # save for later
        image_size = 0
        for section in self.sections:
            if DEBUG: print 'Writing section '+section['Name']
            # update the size of raw data and pointer to raw data information
            section['SizeOfRawData'] = len(section['Data'])
            section['PointerToRawData'] = file_offset
            file_offset += section['SizeOfRawData'] + ((-len(section['Data'])) % self.opt_header['FileAlignment'])
            # find the highest address of virtual memory that we're using
            virtual_size = section['VirtualSize']
            virtual_size = virtual_size + (-virtual_size) % self.opt_header['SectionAlignment']
            virtual_start = section['VirtualAddress']
            if virtual_start % self.opt_header['SectionAlignment']:
                raise Error("The section '"+section['Name']+"' is not correctly alignet in virtual memory")
            image_size = max(image_size, virtual_start+virtual_size)
            # create the section table entry
            section_table += struct.pack('<8sLLLLLLHHL',
                                         section['Name'],
                                         section['VirtualSize'],
                                         section['VirtualAddress'],
                                         section['SizeOfRawData'],
                                         section['PointerToRawData'],
                                         section['PointerToRelocations'],
                                         section['PointerToLinenumbers'],
                                         section['NumberOfRelocations'],
                                         section['NumberOfLinenumbers'],
                                         section['Characteristics'])

        # compile the opt header
        data_directories_data = ''
        # pack the directories
        for virtual_address,size in self.opt_header['DataDirectory']:
            data_directories_data += struct.pack('<LL',virtual_address,size)
        # update the information on the optional header
        self.opt_header['NumberOfRvaAndSizes'] = len(data_directories_data) / 8
        self.opt_header['SizeOfImage'] = image_size
        if DEBUG: print 'Size of Image: '+hex(image_size)
        self.opt_header['SizeOfHeaders'] = size_of_headers
        # put it all together
        self.opt_header_data = struct.pack('<HBBLLLLLLLLLHHHHHHLLLLHHLLLLLL',
                                           self.opt_header['Magic'],
                                           self.opt_header['MajorLinkerVersion'],
                                           self.opt_header['MinorLinkerVersion'],
                                           self.opt_header['SizeOfCode'],
                                           self.opt_header['SizeOfInitializedData'],
                                           self.opt_header['SizeOfUninitializedData'],
                                           self.opt_header['AddressOfEntryPoint'],
                                           self.opt_header['BaseOfCode'],
                                           self.opt_header['BaseOfData'],
                                           self.opt_header['ImageBase'],
                                           self.opt_header['SectionAlignment'],
                                           self.opt_header['FileAlignment'],
                                           self.opt_header['MajorOperatingSystemVersion'],
                                           self.opt_header['MinorOperatingSystemVersion'],
                                           self.opt_header['MajorImageVersion'],
                                           self.opt_header['MinorImageVersion'],
                                           self.opt_header['MajorSubsystemVersion'],
                                           self.opt_header['MinorSubsystemVersion'],
                                           self.opt_header['Reserved1'],
                                           self.opt_header['SizeOfImage'],
                                           self.opt_header['SizeOfHeaders'],
                                           self.opt_header['CheckSum'],
                                           self.opt_header['Subsystem'],
                                           self.opt_header['DllCharacteristics'],
                                           self.opt_header['SizeOfStackReserve'],
                                           self.opt_header['SizeOfStackCommit'],
                                           self.opt_header['SizeOfHeapReserve'],
                                           self.opt_header['SizeOfHeapCommit'],
                                           self.opt_header['LoaderFlags'],
                                           self.opt_header['NumberOfRvaAndSizes'],) + data_directories_data
                                           
        # compile the pe header
        self.pe_header['SizeOfOptionalHeader'] = len(self.opt_header_data)
        self.pe_header_data = struct.pack('<4sHHLLLHH',
                                          self.pe_header['Magic'],
                                          self.pe_header['Machine'],
                                          len(self.sections),
                                          self.pe_header['TimeDateStamp'],
                                          self.pe_header['PointerToSymbolTable'],
                                          self.pe_header['NumberOfSymbols'],
                                          self.pe_header['SizeOfOptionalHeader'],
                                          self.pe_header['Characteristics'],)

        # write the output
        f=open(path,'wb')
        f.write(self.dos_block)
        f.write(self.pe_header_data)
        f.write(self.opt_header_data)
        f.write(section_table)
        if f.tell() > size_of_headers:
            raise Exception('Uh oh!  We overshot our expectation for PE header length: %08X vs. %08X' % (f.tell(),size_of_headers))
        f.seek(size_of_headers,0) # seek ahead if we're not there yet
        for section in self.sections:
            f.seek(section['PointerToRawData'],0)
            f.write(section['Data'][0:section['SizeOfRawData']])
        f.write(self.extra_data)
        f.close()

